@@ -3,24 +3,18 @@
 
 /*---
 esid: sec-array.prototype.reduceright
-es5id: 15.4.4.22-7-6
+es5id: 15.4.4.22-7-5
 description: >
     Array.prototype.reduceRight returns initialValue if 'length' is 0
     and initialValue is present (subclassed Array, length overridden
-    with obj with valueOf)
+    to '0' (type conversion))
 ---*/
 
 foo.prototype = new Array(1, 2, 3);
 
 function foo() {}
 var f = new foo();
-
-var o = {
-  valueOf: function() {
-    return 0;
-  }
-};
-f.length = o;
+f.length = '0';
 
 function cb() {}
 assert.sameValue(f.reduceRight(cb, 1), 1, 'f.reduceRight(cb,1)');
