@@ -34,13 +34,13 @@
 #include <private/qmetaobjectbuilder_p.h>
 #include <private/qqmlcontextdata_p.h>
 #include <QCryptographicHash>
-#include "../../shared/util.h"
+#include <QtQuickTestUtils/private/qmlutils_p.h>
 
 class tst_qqmlpropertycache : public QQmlDataTest
 {
     Q_OBJECT
 public:
-    tst_qqmlpropertycache() {}
+    tst_qqmlpropertycache() : QQmlDataTest(QT_QMLTEST_DATADIR) {}
 
 private slots:
     void properties();
@@ -58,6 +58,7 @@ private slots:
     void metaObjectsForRootElements();
     void derivedGadgetMethod();
     void restrictRegistrationVersion();
+    void rejectOverriddenFinal();
 
 private:
     QQmlEngine engine;
@@ -131,6 +132,7 @@ class BaseObject : public QObject
     Q_PROPERTY(int propertyA READ propertyA NOTIFY propertyAChanged)
     Q_PROPERTY(QString propertyB READ propertyB NOTIFY propertyBChanged)
     Q_PROPERTY(int highVersion READ highVersion WRITE setHighVersion NOTIFY highVersionChanged REVISION(4, 0))
+    Q_PROPERTY(int finalProp READ finalProp CONSTANT FINAL)
 
 public:
     BaseObject(QObject *parent = nullptr) : QObject(parent) {}
@@ -138,6 +140,7 @@ public:
     int propertyA() const { return 0; }
     QString propertyB() const { return QString(); }
     int highVersion() const { return m_highVersion; }
+    int finalProp() const { return 8; }
 
 public Q_SLOTS:
     void slotA() {}
@@ -167,12 +170,14 @@ class DerivedObject : public BaseObject
     Q_PROPERTY(int propertyC READ propertyC NOTIFY propertyCChanged)
     Q_PROPERTY(QString propertyD READ propertyD NOTIFY propertyDChanged)
     Q_PROPERTY(int propertyE READ propertyE NOTIFY propertyEChanged REVISION 1)
+    Q_PROPERTY(int finalProp READ finalProp CONSTANT) // bad!
 public:
     DerivedObject(QObject *parent = nullptr) : BaseObject(parent) {}
 
     int propertyC() const { return 0; }
     QString propertyD() const { return QString(); }
     int propertyE() const { return 0; }
+    int finalProp() const { return 9; }
 
 public Q_SLOTS:
     void slotB() {}
@@ -195,9 +200,8 @@ void tst_qqmlpropertycache::properties()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlPropertyData *data;
+    QQmlRefPointer<QQmlPropertyCache> cache = QQmlPropertyCache::createStandalone(metaObject);
+    const QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "propertyA")));
     QCOMPARE(data->coreIndex(), metaObject->indexOfProperty("propertyA"));
@@ -218,12 +222,11 @@ void tst_qqmlpropertycache::propertiesDerived()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> parentCache(
-                new QQmlPropertyCache(&BaseObject::staticMetaObject),
-                QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlPropertyData *data;
+    QQmlRefPointer<QQmlPropertyCache> parentCache
+            = QQmlPropertyCache::createStandalone(&BaseObject::staticMetaObject);
+    QQmlRefPointer<QQmlPropertyCache> cache
+            = parentCache->copyAndAppend(object.metaObject(), QTypeRevision());
+    const QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "propertyA")));
     QCOMPARE(data->coreIndex(), metaObject->indexOfProperty("propertyA"));
@@ -246,11 +249,11 @@ void tst_qqmlpropertycache::revisionedProperties()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cacheWithoutVersion(new QQmlPropertyCache(metaObject),
-                                                          QQmlRefPointer<QQmlPropertyCache>::Adopt);
+    QQmlRefPointer<QQmlPropertyCache> cacheWithoutVersion(
+                QQmlPropertyCache::createStandalone(metaObject));
     QQmlRefPointer<QQmlPropertyCache> cacheWithVersion(
-                new QQmlPropertyCache(metaObject, QTypeRevision::fromMinorVersion(1)),
-                QQmlRefPointer<QQmlPropertyCache>::Adopt);
+                QQmlPropertyCache::createStandalone(
+                    metaObject, QTypeRevision::fromMinorVersion(1)));
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cacheWithoutVersion, "propertyE")));
@@ -264,9 +267,8 @@ void tst_qqmlpropertycache::methods()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlPropertyData *data;
+    QQmlRefPointer<QQmlPropertyCache> cache(QQmlPropertyCache::createStandalone(metaObject));
+    const QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "slotA")));
     QCOMPARE(data->coreIndex(), metaObject->indexOfMethod("slotA()"));
@@ -300,10 +302,9 @@ void tst_qqmlpropertycache::methodsDerived()
     const QMetaObject *metaObject = object.metaObject();
 
     QQmlRefPointer<QQmlPropertyCache> parentCache(
-                new QQmlPropertyCache(&BaseObject::staticMetaObject),
-                QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
+                QQmlPropertyCache::createStandalone(&BaseObject::staticMetaObject));
+    QQmlRefPointer<QQmlPropertyCache> cache
+            = parentCache->copyAndAppend(object.metaObject(), QTypeRevision {});
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "slotA")));
@@ -337,9 +338,8 @@ void tst_qqmlpropertycache::signalHandlers()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlPropertyData *data;
+    QQmlRefPointer<QQmlPropertyCache> cache(QQmlPropertyCache::createStandalone(metaObject));
+    const QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "onSignalA")));
     QCOMPARE(data->coreIndex(), metaObject->indexOfMethod("signalA()"));
@@ -367,10 +367,9 @@ void tst_qqmlpropertycache::signalHandlersDerived()
     const QMetaObject *metaObject = object.metaObject();
 
     QQmlRefPointer<QQmlPropertyCache> parentCache(
-                new QQmlPropertyCache(&BaseObject::staticMetaObject),
-                QQmlRefPointer<QQmlPropertyCache>::Adopt);
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
-                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
+                QQmlPropertyCache::createStandalone(&BaseObject::staticMetaObject));
+    QQmlRefPointer<QQmlPropertyCache> cache
+            = parentCache->copyAndAppend(object.metaObject(), QTypeRevision{});
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "onSignalA")));
@@ -573,10 +572,28 @@ class TestClassWithClassInfo : public QObject
 #define ARRAY_SIZE(arr) \
     int(sizeof(arr) / sizeof(arr[0]))
 
+template <typename T, typename = void>
+struct SizeofOffsetsAndSizes_helper
+{
+    static constexpr size_t value = sizeof(T::offsetsAndSize); // old moc
+};
+
+template <typename T>
+struct SizeofOffsetsAndSizes_helper<T, std::void_t<decltype(T::offsetsAndSizes)>>
+{
+    static constexpr size_t value = sizeof(T::offsetsAndSizes); // new moc
+};
+
+template <typename T>
+constexpr size_t sizeofOffsetsAndSizes(const T &)
+{
+    return SizeofOffsetsAndSizes_helper<T>::value;
+}
+
 #define TEST_CLASS(Class) \
     QTest::newRow(#Class) \
             << &Class::staticMetaObject << ARRAY_SIZE(qt_meta_data_##Class) \
-            << int(sizeof(qt_meta_stringdata_##Class.offsetsAndSize) / (sizeof(uint) * 2))
+            << int(sizeofOffsetsAndSizes(qt_meta_stringdata_##Class) / (sizeof(uint) * 2))
 
 Q_DECLARE_METATYPE(const QMetaObject*);
 
@@ -654,6 +671,8 @@ void tst_qqmlpropertycache::metaObjectsForRootElements()
 
 void tst_qqmlpropertycache::derivedGadgetMethod()
 {
+    metaObjectsForRootElements();
+
     qmlRegisterTypesAndRevisions<BaseGadget, DerivedGadget, GadgetUser>("Test.PropertyCache", 1);
     QQmlEngine engine;
     QQmlComponent c(&engine, testFileUrl("derivedGadgetMethod.qml"));
@@ -674,4 +693,43 @@ void tst_qqmlpropertycache::restrictRegistrationVersion()
     QVERIFY(c.isError());
 }
 
+void tst_qqmlpropertycache::rejectOverriddenFinal()
+{
+    qmlRegisterTypesAndRevisions<BaseObject>("Test.PropertyCache", 3);
+    QQmlEngine engine;
+
+    QTest::ignoreMessage(QtWarningMsg, QRegularExpression(
+                             "Final member finalProp is overridden in class BaseObject_QML_.* "
+                             "The override won't be used."));
+
+    QQmlComponent c(&engine, testFileUrl("finalProp.qml"));
+    QVERIFY2(!c.isError(), qPrintable(c.errorString()));
+
+    QTest::ignoreMessage(QtWarningMsg, QRegularExpression(
+                             "finalProp.qml:15: TypeError: Property 'finalProp' of object "
+                             "BaseObject_QML_.* is not a function"));
+    QTest::ignoreMessage(QtWarningMsg, QRegularExpression(
+                             "finalProp.qml:11: TypeError: Property 'finalProp' of object "
+                             "BaseObject_QML_.* is not a function"));
+
+    QScopedPointer<QObject> o(c.create());
+    QCOMPARE(o->property("a").toInt(), 8);
+
+    DerivedObject *derived = new DerivedObject(o.data());
+    QTest::ignoreMessage(QtWarningMsg,
+                         "Final member finalProp is overridden in class DerivedObject. "
+                         "The override won't be used.");
+    o->setProperty("obj", QVariant::fromValue(derived));
+    QCOMPARE(derived->finalProp(), 9);
+
+    // rejects override of final property
+    QCOMPARE(o->property("a").toInt(), 8);
+
+    // Cannot override final property with method, either
+    QCOMPARE(o->property("b").toInt(), 8);
+
+    // Cannot call the method overridding a final property
+    QCOMPARE(o->property("c").toInt(), 0);
+}
+
 QTEST_MAIN(tst_qqmlpropertycache)
